#include "model.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

#include <float.h> // for FLT_MAX

#define LINE_SIZE 512

void Model::load(Graphics &gfx, char *file)
{
	char vbo_file[LINE_SIZE];
	char ibo_file[LINE_SIZE];
	char tga_file[LINE_SIZE];
	char normal_file[LINE_SIZE];

	snprintf(vbo_file, LINE_SIZE, "%s.vbo", file);
	snprintf(ibo_file, LINE_SIZE, "%s.ibo", file);
	snprintf(tga_file, LINE_SIZE, "%s.tga", file);
	snprintf(normal_file, LINE_SIZE, "%s_normal.tga", file);

	model_file = get_file(vbo_file);
	if (model_file == NULL)
	{
		printf("Unable to load vertex array %s\n", vbo_file);
		return;
	}

	num_vertex = *((int *)model_file);
	model_vertex_array = (vertex_t *)(model_file + 4);

//	create_box(gfx, aabb);
	make_aabb();

	char *index_file = get_file(ibo_file);
	if (index_file == NULL)
	{
		printf("Unable to load index array %s\n", index_file);
		delete [] model_file;
		return;
	}

	num_index = *((int *)index_file);
	model_index_array = (int *)(index_file + 4);

	model_vertex = gfx.CreateVertexBuffer(model_vertex_array, num_vertex);
	model_index = gfx.CreateIndexBuffer(model_index_array, num_index);
//	delete [] index_file;
//	index_file = NULL;
//	delete [] model_file;
//	model_file = NULL;

}
// copies visual data from model
void Model::clone(Model &model)
{
	for(int i = 0; i < 8; i++)
	{
		aabb[i] = model.aabb[i];
	}
	center = model.center;
	box_index = model.box_index;
	box_vertex = model.box_vertex;
	model_index_array = model.model_index_array;
	model_vertex_array = model.model_vertex_array;
	model_index = model.model_index;
	model_tex = model.model_tex;
	model_vertex = model.model_vertex;
	num_index = model.num_index;
	num_vertex = model.num_vertex;
}

void Model::make_aabb()
{
	aabb[0] = vec3(FLT_MAX, FLT_MAX, FLT_MAX);
	aabb[7] = vec3(-FLT_MAX, -FLT_MAX, -FLT_MAX);
	vec3 sum = vec3();
	for (int i = 0; i < num_vertex; i++)
	{
		sum += model_vertex_array[i].position;

		if (model_vertex_array[i].position.x < aabb[0].x)
			aabb[0].x = model_vertex_array[i].position.x;
		else if (model_vertex_array[i].position.y < aabb[0].y)
			aabb[0].y = model_vertex_array[i].position.y;
		else if (model_vertex_array[i].position.z < aabb[0].z)
			aabb[0].z = model_vertex_array[i].position.z;

		if (model_vertex_array[i].position.x > aabb[7].x)
			aabb[7].x = model_vertex_array[i].position.x;
		else if (model_vertex_array[i].position.y > aabb[7].y)
			aabb[7].y = model_vertex_array[i].position.y;
		else if (model_vertex_array[i].position.z > aabb[7].z)
			aabb[7].z = model_vertex_array[i].position.z;
	}
	center = sum / (float)num_vertex;
	// binary order
	aabb[1] = vec3(aabb[0].x, aabb[0].y, aabb[7].z);
	aabb[2] = vec3(aabb[0].x, aabb[7].y, aabb[0].z);
	aabb[3] = vec3(aabb[0].x, aabb[7].y, aabb[7].z);
	aabb[4] = vec3(aabb[7].x, aabb[0].y, aabb[0].z);
	aabb[5] = vec3(aabb[7].x, aabb[0].y, aabb[7].z);
	aabb[6] = vec3(aabb[7].x, aabb[7].y, aabb[0].z);
}



void Model::render(Graphics &gfx)
{
	gfx.SelectIndexBuffer(model_index);
	gfx.SelectVertexBuffer(model_vertex);
	gfx.DrawArray(PRIM_TRIANGLES, 0, 0, num_index, num_vertex);

}

float *Model::get_matrix(float *matrix)
{
	matrix[0] = morientation.m[0];
	matrix[1] = morientation.m[1];
	matrix[2] = morientation.m[2];
	matrix[3] = 0.0f;

	matrix[4] = morientation.m[3];
	matrix[5] = morientation.m[4];
	matrix[6] = morientation.m[5];
	matrix[7] = 0.0f;

	matrix[8]  = morientation.m[6];
	matrix[9]  = morientation.m[7];
	matrix[10] = morientation.m[8];
	matrix[11] = 0.0f;

	/* matrix rotates around center, but position is arbitrary point
	from which verts are defined */
	vec3 offset = morientation * center;

	matrix[12] = position.x - offset.x;
	matrix[13] = position.y - offset.y;
	matrix[14] = position.z - offset.z;
	matrix[15] = 1.0f;
	return matrix;
}

Model::~Model()
{
	if (model_file != NULL)
		delete [] model_file;
	model_file = NULL;
}


void Model::calc_frustum(matrix4 &clip)
{
//	matrix4	clip;

//	clip = transformation * projection;

	// This will extract the RIGHT side of the frustum
	frustum[0].x = clip.m[ 3] - clip.m[ 0];
	frustum[0].y = clip.m[ 7] - clip.m[ 4];
	frustum[0].z = clip.m[11] - clip.m[ 8];
	frustum[0].w = clip.m[15] - clip.m[12];
	frustum[0].normalize();

	// This will extract the LEFT side of the frustum
	frustum[1].x = clip.m[ 3] + clip.m[ 0];
	frustum[1].y = clip.m[ 7] + clip.m[ 4];
	frustum[1].z = clip.m[11] + clip.m[ 8];
	frustum[1].w = clip.m[15] + clip.m[12];
	frustum[1].normalize();

	// This will extract the BOTTOM side of the frustum
	frustum[2].x = clip.m[ 3] + clip.m[ 1];
	frustum[2].y = clip.m[ 7] + clip.m[ 5];
	frustum[2].z = clip.m[11] + clip.m[ 9];
	frustum[2].w = clip.m[15] + clip.m[13];
	frustum[2].normalize();

	// This will extract the TOP side of the frustum
	frustum[3].x = clip.m[ 3] - clip.m[ 1];
	frustum[3].y = clip.m[ 7] - clip.m[ 5];
	frustum[3].z = clip.m[11] - clip.m[ 9];
	frustum[3].w = clip.m[15] - clip.m[13];
	frustum[3].normalize();

	// This will extract the BACK side of the frustum
	frustum[4].x = clip.m[ 3] - clip.m[ 2];
	frustum[4].y = clip.m[ 7] - clip.m[ 6];
	frustum[4].z = clip.m[11] - clip.m[10];
	frustum[4].w = clip.m[15] - clip.m[14];
	frustum[4].normalize();

	// This will extract the FRONT side of the frustum
	frustum[5].x = clip.m[ 3] + clip.m[ 2];
	frustum[5].y = clip.m[ 7] + clip.m[ 6];
	frustum[5].z = clip.m[11] + clip.m[10];
	frustum[5].w = clip.m[15] + clip.m[14];
	frustum[5].normalize();
}

void Model::getForwardVector(vec3 &forward)
{
	forward.x = morientation.m[6];
	forward.y = morientation.m[7];
	forward.z = morientation.m[8];
}

void Model::get_frame(Frame &frame)
{
	frame.up.x = morientation.m[3];
	frame.up.y = morientation.m[4];
	frame.up.z = morientation.m[5];
	frame.forward.x = -morientation.m[6];
	frame.forward.y = -morientation.m[7];
	frame.forward.z = -morientation.m[8];
	frame.pos = position;
}

Model::Model()
{
	num_vertex = 0;
	num_index = 0;
	normal_tex = 0;
	model_vertex = 0;
	model_index = 0;
	model_tex = 0;
	box_vertex = 0;
	box_index = 0;
	model_file = NULL;
	model_vertex_array = NULL;
	model_index_array = NULL;
	center = vec3();
	rail_trail = false;
	lightning_trail = false;
	blend = false;

	morientation.m[0] = 1.0f;
	morientation.m[1] = 0.0f;
	morientation.m[2] = 0.0f;

	morientation.m[3] = 0.0f;
	morientation.m[4] = 1.0f;
	morientation.m[5] = 0.0f;

	morientation.m[6] = 0.0f;
	morientation.m[7] = 0.0f;
	morientation.m[8] = 1.0f;
}

